'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.queryASTToSqlAST = queryASTToSqlAST;
exports.getGraphQLType = getGraphQLType;
exports.pruneDuplicateSqlDeps = pruneDuplicateSqlDeps;

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _aliasNamespace = require('./aliasNamespace');

var _aliasNamespace2 = _interopRequireDefault(_aliasNamespace);

var _util = require('./util');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function queryASTToSqlAST(resolveInfo, options) {
  const namespace = new _aliasNamespace2.default(options.minify);

  const sqlAST = {};

  const fieldNodes = resolveInfo.fieldNodes || resolveInfo.fieldASTs;
  _assert2.default.equal(fieldNodes.length, 1, 'We thought this would always have a length of 1. FIX ME!!');

  const queryAST = fieldNodes[0];

  const parentType = resolveInfo.parentType;
  getGraphQLType(queryAST, parentType, sqlAST, resolveInfo.fragments, resolveInfo.variableValues, namespace, options);

  pruneDuplicateSqlDeps(sqlAST, namespace);

  return sqlAST;
}

function getGraphQLType(queryASTNode, parentTypeNode, sqlASTNode, fragments, variables, namespace, options) {
  const fieldName = queryASTNode.name.value;

  if (fieldName.slice(0, 2) === '__') {
    sqlASTNode.type = 'noop';
    return;
  }

  let field = parentTypeNode._fields[fieldName];
  if (!field) {
    throw new Error(`The field "${fieldName}" is not in the ${parentTypeNode.name} type.`);
  }

  let grabMany = false;

  let gqlType = stripNonNullType(field.type);

  if (queryASTNode.arguments.length) {
    const args = sqlASTNode.args = {};
    for (let arg of queryASTNode.arguments) {
      args[arg.name.value] = parseArgValue(arg.value, variables);
    }
  }

  if (gqlType.constructor.name === 'GraphQLList') {
    gqlType = gqlType.ofType;
    grabMany = true;
  }

  if (gqlType.constructor.name === 'GraphQLObjectType' && gqlType._fields.edges && gqlType._fields.pageInfo) {
    grabMany = true;

    const stripped = stripRelayConnection(field, queryASTNode);

    gqlType = stripped.gqlType;
    queryASTNode = stripped.queryASTNode;

    if (field.sqlPaginate) {
      if (options.dialect !== 'pg') {
        throw new Error('Cannot do SQL pagination for connections with this SQL dialect');
      }
      sqlASTNode.paginate = true;

      getSortColumns(field, sqlASTNode);
    }
  } else {
    if (field.sqlPaginate) {
      throw new Error(`To paginate the ${gqlType.name} type, it must be a GraphQLObjectType that fulfills the relay spec. The type must have a "pageInfo" and "edges" field. https://facebook.github.io/relay/graphql/connections.htm`);
    }
  }

  const config = gqlType._typeConfig;

  if (gqlType.constructor.name === 'GraphQLObjectType' && config.sqlTable) {
    handleTable(sqlASTNode, queryASTNode, field, gqlType, fragments, variables, namespace, grabMany, options);
  } else if (field.sqlExpr) {
    sqlASTNode.type = 'expression';
    sqlASTNode.sqlExpr = field.sqlExpr;
    sqlASTNode.fieldName = field.name;
    sqlASTNode.as = namespace.generate('column', field.name);
  } else if (field.sqlColumn || !field.resolve) {
    sqlASTNode.type = 'column';
    sqlASTNode.name = field.sqlColumn || field.name;
    sqlASTNode.fieldName = field.name;
    sqlASTNode.as = namespace.generate('column', sqlASTNode.name);
  } else if (field.sqlDeps) {
    sqlASTNode.type = 'columnDeps';
    sqlASTNode.names = field.sqlDeps;
  } else {
    sqlASTNode.type = 'noop';
  }
}

function handleTable(sqlASTNode, queryASTNode, field, gqlType, fragments, variables, namespace, grabMany, options) {
  const config = gqlType._typeConfig;

  sqlASTNode.type = 'table';
  sqlASTNode.name = config.sqlTable;

  sqlASTNode.as = namespace.generate('table', field.name);

  sqlASTNode.fieldName = field.name;
  sqlASTNode.grabMany = grabMany;

  if (field.where) {
    sqlASTNode.where = field.where;
  }
  if (field.sqlJoin) {
    sqlASTNode.sqlJoin = field.sqlJoin;
  }
  if (field.joinTable) {
    sqlASTNode.sqlJoins = field.sqlJoins;
    sqlASTNode.joinTable = field.joinTable;
    sqlASTNode.joinTableAs = namespace.generate('table', field.joinTable);
  }

  const children = sqlASTNode.children = [];

  handleUniqueKey(config, children, namespace);

  if (sqlASTNode.paginate) {
    handleColumnsRequiredForPagination(sqlASTNode, namespace);
  }

  if (queryASTNode.selectionSet) {
    handleSelections(children, queryASTNode.selectionSet.selections, gqlType, fragments, variables, namespace, options);
  }
}

function handleSelections(children, selections, gqlType, fragments, variables, namespace, options) {
  for (let selection of selections) {
    switch (selection.kind) {
      case 'Field':
        const newNode = {};
        children.push(newNode);
        getGraphQLType(selection, gqlType, newNode, fragments, variables, namespace, options);
        break;

      case 'InlineFragment':
        {
          const selectionNameOfType = selection.typeCondition.name.value;
          const sameType = selectionNameOfType === gqlType.name;
          const interfaceType = gqlType._interfaces.map(iface => iface.name).indexOf(selectionNameOfType) >= 0;
          if (sameType || interfaceType) {
            handleSelections(children, selection.selectionSet.selections, gqlType, fragments, variables, namespace, options);
          }
        }
        break;

      case 'FragmentSpread':
        {
          const fragmentName = selection.name.value;
          const fragment = fragments[fragmentName];

          const fragmentNameOfType = fragment.typeCondition.name.value;
          const sameType = fragmentNameOfType === gqlType.name;
          const interfaceType = gqlType._interfaces.map(iface => iface.name).indexOf(fragmentNameOfType) >= 0;
          if (sameType || interfaceType) {
            handleSelections(children, fragment.selectionSet.selections, gqlType, fragments, variables, namespace, options);
          }
        }
        break;
      default:
        throw new Error('Unknown selection kind: ' + selection.kind);
    }
  }
}

function handleUniqueKey(config, children, namespace) {
  if (!config.uniqueKey) {
    throw new Error(`You must specify the "uniqueKey" on the GraphQLObjectType definition of ${config.sqlTable}`);
  }
  if (typeof config.uniqueKey === 'string') {
    children.push({
      type: 'column',
      name: config.uniqueKey,
      fieldName: config.uniqueKey,
      as: namespace.generate('column', config.uniqueKey)
    });
  } else if (Array.isArray(config.uniqueKey)) {
    const clumsyName = config.uniqueKey.join('#');
    children.push({
      type: 'composite',
      name: config.uniqueKey,
      fieldName: clumsyName,
      as: namespace.generate('column', clumsyName)
    });
  }
}

function handleColumnsRequiredForPagination(sqlASTNode, namespace) {
  if (sqlASTNode.sortKey) {
    for (let column of (0, _util.wrap)(sqlASTNode.sortKey.key)) {
      const newChild = {
        type: 'column',
        name: column,
        fieldName: column,
        as: namespace.generate('column', column)
      };

      if (sqlASTNode.joinTable) {
        newChild.fromOtherTable = sqlASTNode.joinTableAs;
      }
      sqlASTNode.children.push(newChild);
    }
  } else if (sqlASTNode.orderBy) {
    const newChild = {
      type: 'column',
      name: '$total',
      fieldName: '$total',
      as: namespace.generate('column', '$total')
    };
    if (sqlASTNode.joinTable) {
      newChild.fromOtherTable = sqlASTNode.joinTableAs;
    }
    sqlASTNode.children.push(newChild);
  }
}

function stripRelayConnection(field, queryASTNode) {
  const gqlType = field.type._fields.edges.type.ofType._fields.node.type;

  const args = queryASTNode.arguments;

  const edges = queryASTNode.selectionSet.selections.find(selection => selection.name.value === 'edges');
  if (edges) {
    queryASTNode = edges.selectionSet.selections.find(selection => selection.name.value === 'node') || {};
  } else {
    queryASTNode = {};
  }

  queryASTNode.arguments = args;
  return { gqlType, queryASTNode };
}

function stripNonNullType(type) {
  return type.constructor.name === 'GraphQLNonNull' ? type.ofType : type;
}

function pruneDuplicateSqlDeps(sqlAST, namespace) {
  const deps = new Set();
  const children = sqlAST.children;

  for (let i = children.length - 1; i >= 0; i--) {
    const child = children[i];
    if (child.type === 'columnDeps') {
      child.names.forEach(name => deps.add(name));
      children.splice(i, 1);
    } else if (child.type === 'table') {
      pruneDuplicateSqlDeps(child, namespace);
    }
  }

  const newNode = {
    type: 'columnDeps',
    names: {}
  };
  deps.forEach(name => {
    newNode.names[name] = namespace.generate('column', name);
  });
  children.push(newNode);
}

function parseArgValue(value, variableValues) {
  if (value.kind === 'Variable') {
    const variableName = value.name.value;
    return variableValues[variableName];
  }

  let primitive = value.value;

  if (value.kind === 'IntValue') {
    primitive = parseInt(primitive);
  }
  return primitive;
}

function getSortColumns(field, sqlASTNode) {
  if (field.sortKey) {
    if (typeof field.sortKey === 'function') {
      sqlASTNode.sortKey = field.sortKey(sqlASTNode.args);
    } else {
      sqlASTNode.sortKey = field.sortKey;
    }
  } else if (field.orderBy) {
    if (typeof field.orderBy === 'function') {
      sqlASTNode.orderBy = field.orderBy(sqlASTNode.args);
    } else {
      sqlASTNode.orderBy = field.orderBy;
    }
  }
}